from errors import InstallError
import sys, os
import argparse
import logging
import copy
import json
import subprocess
import getpass
import time

# Yaml
import yaml
from yaml import Loader, SafeLoader

# Globus SDK
# import globus_sdk
# from globus_sdk import AuthClient, AccessTokenAuthorizer
# from globus_sdk import ConfidentialAppAuthClient

# OIDC Server
# from globus_oidc.config import settings

# Mako
# from mako.template import Template


def render_template(config: dict, srcfilename: str, dstfilename: str):
    if not os.path.isfile(srcfilename):
        raise InstallError("Render Template", "Cannot find config file `{}`".format(srcfilename))
    with open(srcfilename) as f:
        template = Template(f.read())
        with open(dstfilename, "w") as fp:
            fp.write(template.render(config))


def load_config(filename: str) -> dict:
    """
    Get the YAML configuration from a provided file location.
    :param filename: The filename.
    :return: Configuration as a dictionary.
    """
    if os.path.isfile(filename):
        with open(filename) as f:
            return yaml.safe_load(f)
    else:
        raise InstallError("Load Config", "Cannot find config file `{}`".format(filename))


def initial_validation(config: dict):
    """
    Performs the initial validation before installing.
    :return: Nothing, exception thrown if any issues detected.
    """

    # TODO
    # Check that versions match
    if not config['oidc_server']['version'].isspace():
        # We have a bound configuration version.
        # Load settings
        app_settings = settings()
        if config['oidc_server']['version'] != app_settings['config']['VERSION']:
            raise InstallError("Initial Validation",
                               "Conflict between existing installed version '{}' and installing version '{}'. Please install the same version as the other nodes."
                               .format(config['oidc_server']['version'],app_settings.VERSION))

    # Check that we're root
    if os.geteuid() != 0:
        raise InstallError("Initial Validation",
                           "This installer is not running as root, please rerun as root.")

    # Check that current user is part of the shadow group


def preregister_validation(config: dict, client: dict):
    """
    Performs validation before registering with Globus Auth. Check that we can see the well-known endpoint from local.
    :param config:
    :param client:
    :return: Nothing, exception thrown if any issues detected.
    """
    log.info("Validating that the OIDC .well-known file can be accessed from the FQDN...")

    url = "https://{}/.well-known/openid-configuration".format(config['auth']['domain'])
    found = False
    while not found:
        log.info("Attempting to retrieve '{}'...".format(url))
        time.sleep(5)

    pass


def pam_validation(config: dict):
    """
    Performs a pamtester test to validate that the service would work if you log in.
    :param config:
    :return:
    """
    service = config['oidc_server']['pam_service']
    user = config['USER']
    print("PAM Validation: Please log in as the user to confirm that PAM is functioning correctly.\r\n")
    print("   PAM Service: {}".format(config['oidc_server']['pam_service']))
    # print("User [{}]: ".format(user))
    # userin = input()
    # if not userin.strip().isspace():
    #     user = userin
    rc = subprocess.call("sudo pamtester {} {} authenticate".format(service, user), shell=True)
    log.info("PAM Return code = {}".format(rc))
    if rc != 0:
        raise InstallError("PAM Validation", "Unable to successfully login via PAM service, please fix the PAM service and rerun the installer.")
    else:
        log.info("Successfully logged in via PAM!")


def get_auth_client(auth: dict) -> dict:
    """
    Get the auth client (including domains and project).
    :param auth: The auth configuration dictionary.
    :return: Auth client including domains and project.
    """
    # Create the auth
    authclient = ConfidentialAppAuthClient(
        auth['client_id'], auth['client_secret'])

    # Get the client
    resp = authclient.get("/v2/api/clients/{}".format(auth['client_id']))
    if resp.http_status != 200:
        raise InstallError("Get Auth Client", "Unable to retrieve client, Globus Auth returned status code {}.".format(resp.http_status))
    # Why does a GET of a single item return a dictionary?
    if 'client' not in resp.data:
        raise InstallError("Get Auth Client", "Unable to retrieve client, returned body did not contain a `client` entry.")
    client = resp.data['client']

    # TODO: May have to recursively seek parent projects to find the project ID.
    # TODO: Expose project endpoint in auth.
    # Get the project
    # resp = authclient.get("/v2/web/developers/projects/{}/admins".format(client['project']))
    # if resp.http_status != 200:
    #     raise Exception("Unable to retrieve client's project")
    # client.project = resp.data

    log.debug("Retrieved client: \r\n{}".format(client))
    log.info("Successfully retrieved client '{}'.".format(client['id']))

    return client


def update_auth_redirect_urls(auth: dict, client: dict):
    """
    If the redirect URL is not listed in the client, it is created and updated.
    :param auth: The auth configuration.
    :param client: The current client.
    :return: Nothing, exception if any issue.
    """
    redirect = "https://{}/p/authenticate/callback".format(auth['domain'])
    if redirect not in client["redirect_urls"]:
        raise InstallError("Get Auth Client",
                           "Redirect URL '{}' not found in client '{}', please update the client to include it.".format(redirect, client['id']))
        # TODO: Don't fail here, if it doesn't exist, register it.
        log.debug("Redirect URL '{}' not found in client '{}', registering...").format(redirect, client['id'])
        client["redirect_urls"].append(redirect)

        # TODO: Send this client update to auth.
    else:
        log.debug("Redirect URL '{}' already in client '{}'.").format(redirect, client['id'])


def user_config(origconfig: dict) -> dict:
    """
    Confirm or edit the configuration fields.
    :param origconfig: The original configuration
    :return: The updated configuration.
    """
    config = copy.deepcopy(origconfig)

    print("\r\nCurrent configuration for the OIDC Server:-------\r\n{}\r\n-----\r\n"
          .format(json.dumps(config, indent=2)))
    res = "a"
    while res.strip().lower() not in ['y', 'n']:
        res = input("Would you like to modify the configuration before registering this server? [Y/n] > ")
    if res.strip().lower() == "n":
        log.debug("User accepted default configuration.")
        return config

    # TODO: Let user configure settings.

    return config

def install_uwsgi(config: dict):
    """
    Set up the UWSGI service.
    :return:
    """
    # Render and install the UWSGI service
    # jinja_render_template(config, "./install/etc/globus-oidc.service.jinja", "./globus-oidc.service")
    render_template(config, "./install/etc/globus-oidc.service.mako", "/etc/systemd/system/globus-oidc.service")
    # Run the UWSGI installer
    rc = subprocess.call("./install/bin/config_uwsgi.sh", shell=True)


def install_apache(config: dict):
    # Render and install the Apache service
    # jinja_render_template(config, "./install/etc/globus-oidc.conf.jinja", "./globus-oidc.conf")
    render_template(config, "./install/etc/globus-oidc.conf.mako", "/etc/apache2/sites-available/globus-oidc.conf")
    # Run the Apache installer
    rc = subprocess.call("./install/bin/config_apache.sh", shell=True)


def register_idp(config: dict):
    # Required Inputs -- will need to be passed to the script
    # OIDC server needs to be running and serving this document:
    autodiscovery_url = 'https://oidc.{}/.well-known/openid-configuration'.format(config['auth']['domain'])
    # /Required Inputs

    params = {
        'discovery_url': autodiscovery_url,
        'display_name': config['oidc_server']['display_name'],
        'client_credentials': {
            # The creds for Auth to talk to the OIDC server.
            # TODO: these will need to be generated for each install
            'client_id': config['auth']['client_id'],
            'client_secret': config['auth']['client_secret'],
        },
        'attribute_map': {
            # For installs that are able to provide it, this should also contain
            # display name and email
            'username': 'sub',
            'immutable_id': 'sub',
        },
        'contact_email': config['oidc_server']['support_contact'],
        'contact_name': config['oidc_server']['support_email'],
    }

    # Create the auth
    authclient = ConfidentialAppAuthClient(
        config['auth']['client_id'], config['auth']['client_secret'])

    # Get the client
    resp = authclient.post("/v2/api/identity_providers", json=params)
    if resp.http_status != 200:
        raise InstallError("Create IdP",
                           "Unable to create IdP, Globus Auth returned status code {}.".format(resp.http_status))


def main(argv):
    """
    This is the entrypoint for the standalong installer.
    :param argv:
    :return:
    """
    try:
        """ Runs the installation script.
        """
        parser = argparse.ArgumentParser()
        parser.add_argument('--config', required=True)
        parser.add_argument('--pyenv', required=True)
        parser.add_argument('--user', required=False)
        args = parser.parse_args()

        # Load the AppSync configuration
        if not os.path.isfile(args.config):
            raise InstallError("Initial Validation",
                               "Cannot find the config file {}.".format(args.config))

        config = load_config(args.config)
        config["USER"] = args.user if args.user else getpass.getuser()
        config["ENVPATH"] = args.pyenv
        config["INSTALLPATH"] = os.getcwd()

        print("""----------------\r\nGlobus OIDC Installation
    Configuration: {}
    Python environment: {}
    User: {}
----------------""".format(args.config, args.pyenv, config["USER"]))

        # Tell Globus SDK what environment we're in.
        os.environ["GLOBUS_SDK_ENVIRONMENT"] = config['auth']['env']

        # Do an initial validation of the version etc.
        initial_validation(config)

        # Check that a user can log in via PAM if we use this service.
        pam_validation(config)

        # Get the client
        client = get_auth_client(config['auth'])
        if len(client['fqdns']) == 0:
            raise InstallError("Validate Config",
                               "Auth client does not have any registered FQDN's. Please register at least one FQDN before installing.")
        config_domain = config['auth']['domain'].strip()
        # If no domain is provided, pick the first from the client
        if config_domain.isspace():
            config_domain = client['fqdns'][0]
            log.info("No domain in configuration, selecting '{}' from client.".format(config_domain))

        # Update parameters
        config['auth']['domain'] = config_domain
        # TODO: generate ID

        # Get the user confirmation on the parameters
        config = user_config(config)

        # Validate parameters
        # Assert that FQDN exists and is in the list.
        if config_domain not in client['fqdns']:
            raise InstallError("Validate Config", "OIDC domain '{}' is not a FQDN for the client. Please register the domain before installing".format(config_domain))

        # Install UWSGI
        install_uwsgi(config)

        # Install Apache
        install_apache(config)

        # Do preregistration validation
        preregister_validation(config, client)

    except InstallError as ex:
        log.info("""Unable to install due to installation error.
            Step: {}
            Message: {}
            -----
            Please rectify the issue and try installing again.""".format(ex.step, ex.message))
    except Exception as ex:
        log.info("""Unable to install due to unexpected exception.
            Message: {}
            -----
            Please rectify the issue or contact Globus for troubleshooting assistance.""".format(ex.args))